@@ -38,8 +38,10 @@
 #include <nghttp2/nghttp2.h>
 
 #include "shrpx_io_control.h"
+#include "shrpx_log_config.h"
 #include "http2.h"
 #include "memchunk.h"
+#include "allocator.h"
 
 using namespace nghttp2;
 
@@ -48,21 +50,24 @@ namespace shrpx {
 class Upstream;
 class DownstreamConnection;
 struct BlockedLink;
+struct DownstreamAddrGroup;
+struct DownstreamAddr;
 
 class FieldStore {
 public:
-  FieldStore(size_t headers_initial_capacity)
+  FieldStore(BlockAllocator &balloc, size_t headers_initial_capacity)
       : content_length(-1),
+        balloc_(balloc),
         buffer_size_(0),
         header_key_prev_(false),
         trailer_key_prev_(false) {
     headers_.reserve(headers_initial_capacity);
   }
 
-  const Headers &headers() const { return headers_; }
-  const Headers &trailers() const { return trailers_; }
+  const HeaderRefs &headers() const { return headers_; }
+  const HeaderRefs &trailers() const { return trailers_; }
 
-  Headers &headers() { return headers_; }
+  HeaderRefs &headers() { return headers_; }
 
   const void add_extra_buffer_size(size_t n) { buffer_size_ += n; }
   size_t buffer_size() const { return buffer_size_; }
@@ -72,18 +77,20 @@ public:
   // Returns pointer to the header field with the name |name|.  If
   // multiple header have |name| as name, return last occurrence from
   // the beginning.  If no such header is found, returns nullptr.
-  // This function must be called after headers are indexed
-  const Headers::value_type *header(int32_t token) const;
-  Headers::value_type *header(int32_t token);
+  const HeaderRefs::value_type *header(int32_t token) const;
+  HeaderRefs::value_type *header(int32_t token);
   // Returns pointer to the header field with the name |name|.  If no
   // such header is found, returns nullptr.
-  const Headers::value_type *header(const StringRef &name) const;
+  const HeaderRefs::value_type *header(const StringRef &name) const;
 
-  void add_header_lower(const StringRef &name, const StringRef &value,
-                        bool no_index);
   void add_header_token(const StringRef &name, const StringRef &value,
                         bool no_index, int32_t token);
 
+  // Adds header field name |name|.  First, the copy of header field
+  // name pointed by name.c_str() of length name.size() is made, and
+  // stored.
+  void alloc_add_header_name(const StringRef &name);
+
   void append_last_header_key(const char *data, size_t len);
   void append_last_header_value(const char *data, size_t len);
 
@@ -96,11 +103,14 @@ public:
   // Empties headers.
   void clear_headers();
 
-  void add_trailer_lower(const StringRef &name, const StringRef &value,
-                         bool no_index);
   void add_trailer_token(const StringRef &name, const StringRef &value,
                          bool no_index, int32_t token);
 
+  // Adds trailer field name |name|.  First, the copy of trailer field
+  // name pointed by name.c_str() of length name.size() is made, and
+  // stored.
+  void alloc_add_trailer_name(const StringRef &name);
+
   void append_last_trailer_key(const char *data, size_t len);
   void append_last_trailer_value(const char *data, size_t len);
 
@@ -110,10 +120,11 @@ public:
   int64_t content_length;
 
 private:
-  Headers headers_;
+  BlockAllocator &balloc_;
+  HeaderRefs headers_;
   // trailer fields.  For HTTP/1.1, trailer fields are only included
   // with chunked encoding.  For HTTP/2, there is no such limit.
-  Headers trailers_;
+  HeaderRefs trailers_;
   // Sum of the length of name and value in headers_ and trailers_.
   // This could also be increased by add_extra_buffer_size() to take
   // into account for request URI in case of HTTP/1.x request.
@@ -123,8 +134,8 @@ private:
 };
 
 struct Request {
-  Request()
-      : fs(16),
+  Request(BlockAllocator &balloc)
+      : fs(balloc, 16),
         recv_body_length(0),
         unconsumed_body_length(0),
         method(-1),
@@ -142,19 +153,21 @@ struct Request {
   }
 
   FieldStore fs;
+  // Timestamp when all request header fields are received.
+  std::shared_ptr<Timestamp> tstamp;
   // Request scheme.  For HTTP/2, this is :scheme header field value.
   // For HTTP/1.1, this is deduced from URI or connection.
-  std::string scheme;
+  StringRef scheme;
   // Request authority.  This is HTTP/2 :authority header field value
   // or host header field value.  We may deduce it from absolute-form
   // HTTP/1 request.  We also store authority-form HTTP/1 request.
   // This could be empty if request comes from HTTP/1.0 without Host
   // header field and origin-form.
-  std::string authority;
+  StringRef authority;
   // Request path, including query component.  For HTTP/1.1, this is
   // request-target.  For HTTP/2, this is :path header field value.
   // For CONNECT request, this is empty.
-  std::string path;
+  StringRef path;
   // the length of request body received so far
   int64_t recv_body_length;
   // The number of bytes not consumed by the application yet.
@@ -179,14 +192,15 @@ struct Request {
 };
 
 struct Response {
-  Response()
-      : fs(32),
+  Response(BlockAllocator &balloc)
+      : fs(balloc, 32),
         recv_body_length(0),
         unconsumed_body_length(0),
         http_status(0),
         http_major(1),
         http_minor(1),
-        connection_close(false) {}
+        connection_close(false),
+        headers_only(false) {}
 
   void consume(size_t len) {
     assert(unconsumed_body_length >= len);
@@ -203,6 +217,10 @@ struct Response {
   unsigned int http_status;
   int http_major, http_minor;
   bool connection_close;
+  // true if response only consists of HEADERS, and it bears
+  // END_STREAM.  This is used to tell Http2Upstream that it can send
+  // response with single HEADERS with END_STREAM flag only.
+  bool headers_only;
 };
 
 class Downstream {
@@ -234,7 +252,7 @@ public:
   // Returns true if upgrade (HTTP Upgrade or CONNECT) is succeeded.
   // This should not depend on inspect_http1_response().
   void check_upgrade_fulfilled();
-  // Returns true if the upgrade is succeded as a result of the call
+  // Returns true if the upgrade is succeeded as a result of the call
   // check_upgrade_fulfilled().  HTTP/2 Upgrade is excluded.
   bool get_upgraded() const;
   // Inspects HTTP/2 request.
@@ -245,7 +263,7 @@ public:
   // Returns true if the request is HTTP Upgrade for HTTP/2
   bool get_http2_upgrade_request() const;
   // Returns the value of HTTP2-Settings request header field.
-  const std::string &get_http2_settings() const;
+  StringRef get_http2_settings() const;
 
   // downstream request API
   const Request &request() const { return req_; }
@@ -258,7 +276,7 @@ public:
   void crumble_request_cookie(std::vector<nghttp2_nv> &nva);
   // Assembles request cookies.  The opposite operation against
   // crumble_request_cookie().
-  std::string assemble_request_cookie() const;
+  StringRef assemble_request_cookie();
 
   void
   set_request_start_time(std::chrono::high_resolution_clock::time_point time);
@@ -272,8 +290,9 @@ public:
   // Validates that received request body length and content-length
   // matches.
   bool validate_request_recv_body_length() const;
-  void set_request_downstream_host(std::string host);
+  void set_request_downstream_host(const StringRef &host);
   bool expect_response_body() const;
+  bool expect_response_trailer() const;
   enum {
     INITIAL,
     HEADER_COMPLETE,
@@ -295,7 +314,12 @@ public:
   DefaultMemchunks *get_request_buf();
   void set_request_pending(bool f);
   bool get_request_pending() const;
+  void set_request_header_sent(bool f);
+  bool get_request_header_sent() const;
   // Returns true if request is ready to be submitted to downstream.
+  // When sending pending request, get_request_pending() should be
+  // checked too because this function may return true when
+  // get_request_pending() returns false.
   bool request_submission_ready() const;
 
   // downstream response API
@@ -303,7 +327,7 @@ public:
   Response &response() { return resp_; }
 
   // Rewrites the location response header field.
-  void rewrite_location_response_header(const std::string &upstream_scheme);
+  void rewrite_location_response_header(const StringRef &upstream_scheme);
 
   bool get_chunked_response() const;
   void set_chunked_response(bool f);
@@ -324,6 +348,9 @@ public:
   // True if the response is non-final (1xx status code).  Note that
   // if connection was upgraded, 101 status code is treated as final.
   bool get_non_final_response() const;
+  // True if protocol version used by client supports non final
+  // response.  Only HTTP/1.1 and HTTP/2 clients support it.
+  bool supports_non_final_response() const;
   void set_expect_final_response(bool f);
   bool get_expect_final_response() const;
 
@@ -373,6 +400,30 @@ public:
 
   DefaultMemchunks pop_response_buf();
 
+  BlockAllocator &get_block_allocator();
+
+  void add_rcbuf(nghttp2_rcbuf *rcbuf);
+
+  void
+  set_downstream_addr_group(const std::shared_ptr<DownstreamAddrGroup> &group);
+  void set_addr(const DownstreamAddr *addr);
+
+  const DownstreamAddr *get_addr() const;
+
+  void set_accesslog_written(bool f);
+
+  // Finds affinity cookie from request header fields.  The name of
+  // cookie is given in |name|.  If an affinity cookie is found, it is
+  // assigned to a member function, and is returned.  If it is not
+  // found, or is malformed, returns 0.
+  uint32_t find_affinity_cookie(const StringRef &name);
+  // Set |h| as affinity cookie.
+  void renew_affinity_cookie(uint32_t h);
+  // Returns affinity cookie to send.  If it does not need to be sent,
+  // for example, because the value is retrieved from a request header
+  // field, returns 0.
+  uint32_t get_affinity_cookie_to_send() const;
+
   enum {
     EVENT_ERROR = 0x1,
     EVENT_TIMEOUT = 0x2,
@@ -392,6 +443,10 @@ public:
   int64_t response_sent_body_length;
 
 private:
+  BlockAllocator balloc_;
+
+  std::vector<nghttp2_rcbuf *> rcbufs_;
+
   Request req_;
   Response resp_;
 
@@ -400,7 +455,7 @@ private:
   // host we requested to downstream.  This is used to rewrite
   // location header field to decide the location should be rewritten
   // or not.
-  std::string request_downstream_host_;
+  StringRef request_downstream_host_;
 
   DefaultMemchunks request_buf_;
   DefaultMemchunks response_buf_;
@@ -414,8 +469,12 @@ private:
   Upstream *upstream_;
   std::unique_ptr<DownstreamConnection> dconn_;
 
-  // only used by HTTP/2 or SPDY upstream
+  // only used by HTTP/2 upstream
   BlockedLink *blocked_link_;
+  // The backend address used to fulfill this request.  These are for
+  // logging purpose.
+  std::shared_ptr<DownstreamAddrGroup> group_;
+  const DownstreamAddr *addr_;
   // How many times we tried in backend connection
   size_t num_retry_;
   // The stream ID in frontend connection
@@ -427,11 +486,13 @@ private:
   int32_t downstream_stream_id_;
   // RST_STREAM error_code from downstream HTTP2 connection
   uint32_t response_rst_stream_error_code_;
+  // An affinity cookie value.
+  uint32_t affinity_cookie_;
   // request state
   int request_state_;
   // response state
   int response_state_;
-  // only used by HTTP/2 or SPDY upstream
+  // only used by HTTP/2 upstream
   int dispatch_state_;
   // true if the connection is upgraded (HTTP Upgrade or CONNECT),
   // excluding upgrade to HTTP/2.
@@ -446,6 +507,12 @@ private:
   // has not been established or should be checked before use;
   // currently used only with HTTP/2 connection.
   bool request_pending_;
+  // true if downstream request header is considered to be sent.
+  bool request_header_sent_;
+  // true if access.log has been written.
+  bool accesslog_written_;
+  // true if affinity cookie is generated for this request.
+  bool new_affinity_cookie_;
 };
 
 } // namespace shrpx
